package com.me.common.ecs;

import com.badlogic.gdx.utils.Array;
import com.me.common.ecs.event.Event;
import com.me.common.ecs.event.Listener;

public class Engine {

    private final EntityManager entityManager;
    private final ComponentBag[] components;
    private final Array<BaseSystem> systems;

    private final ListenerRegistry listenerRegistry;

    public Engine() {
        this.entityManager = new EntityManager(this);
        this.components = new ComponentBag[Long.SIZE];
        this.systems = new Array<>();

        this.listenerRegistry = new ListenerRegistry();
    }

    /**
     * Register a Component class.
     * @param clazz
     */
    public void registerComponentClass(Class<? extends Component> clazz) {
        ComponentType type = ComponentType.registerComponentType(clazz);
        components[type.id] = new ComponentBag();
    }

    /**
     * Register a BaseSystem.
     * @param system
     */
    public void registerSystem(BaseSystem system) {
        systems.add(system);
    }

    /**
     * Register an event Listener.
     * @param listener
     */
    public void registerListener(Listener listener) {
        listenerRegistry.registerListener(listener);
    }

    /**
     * Process all systems
     * @param dt the time (in seconds) passed since the last update
     */
    public void update(float dt) {
        entityManager.update();

        for (BaseSystem system : systems) {
            system.preProcess();
            system.process(dt);
            system.postProcess();
        }
    }

    /**
     * @return all entities known by the engine - both those active and inactive.
     */
    public Array<Entity> getEntities() {
        return entityManager.getEntities();
    }

    /**
     * @return all currently registered systems.
     */
    public Array<BaseSystem> getSystems() {
        return systems;
    }

    /**
     * @return a new Entity object which can have fresh components added to it
     */
    public Entity createEntity() {
        return entityManager.create();
    }

    /**
     * Calls the given event, dispatching it to all registered Listeners that have an EventHandler
     * for this particular event.
     * @param event the event to call
     */
    public void callEvent(Event event) {
        listenerRegistry.callEvent(event);
    }

    void refreshEntity(Entity entity) {
        entityManager.queueRefresh(entity);
    }

    void removeAllEntityComponents(int entityId) {
        for (ComponentBag bag : components) {
            if (bag == null) {
                // The rest of the array should also be null given that we never nullify
                // an index after it has been set, and they are set sequentially. So, break.
                break;
            }
            bag.remove(entityId);
        }
    }

    void addEntityComponent(Entity entity, Component component) {
        ComponentType type = ComponentType.getComponentType(component.getClass());
        components[type.id].insert(entity.id, component);
        entity.addComponentType(type);
    }


    void removeEntityComponent(Entity entity, Component component) {
        ComponentType type = ComponentType.getComponentType(component.getClass());
        components[type.id].remove(entity.id);
        entity.removeComponentType(type);
    }

    <T extends Component> T getEntityComponent(Entity entity, Class<T> clazz) {
        return clazz.cast(getEntityComponent(entity, ComponentType.getComponentType(clazz)));
    }

    Component getEntityComponent(Entity entity, ComponentType type) {
        return components[type.id].get(entity.id);
    }

}
