package com.me.common.ecs;

public final class Entity {

    private static int nextId = 0;
    private static long nextUniqueId = 0;

    private final Engine engine;
    final int id;

    long componentBits;
    long systemEnabledBits;
    boolean pendingRefresh;

    boolean active;
    boolean removed;

    private long uniqueId;

    protected Entity(Engine engine) {
        this.engine = engine;
        this.id = nextId++;
    }

    public int getId() {
        return id;
    }

    public long getUniqueId() {
        return uniqueId;
    }

    public Engine getEngine() {
        return engine;
    }

    public boolean isActive() {
        return active;
    }

    public boolean isRemoved() {
        return removed;
    }

    public void activate() {
        if (!active) {
            active = true;
            refresh();
        }
    }

    public void deactivate() {
        if (active) {
            active = false;
            refresh();
        }
    }

    public void remove() {
        if (!removed) {
            removed = true;
            active = false;
            refresh();
        }
    }

    public void addComponent(Component component) {
        engine.addEntityComponent(this, component);
        refresh();
    }

    public void removeComponent(Component component) {
        engine.removeEntityComponent(this, component);
        refresh();
    }

    private void refresh() {
        if (!pendingRefresh) {
            engine.refreshEntity(this);
            pendingRefresh = true;
        }
    }

    public boolean hasComponent(Class<? extends Component> clazz) {
        return ComponentType.getComponentType(clazz).isInBits(componentBits);
    }

    public <T extends Component> T getComponent(Class<T> clazz) {
        return engine.getEntityComponent(this, clazz);
    }

    // Internal getComponent method accessed only by ComponentMapper
    Component getComponent(ComponentType type) {
        return engine.getEntityComponent(this, type);
    }

    void addComponentType(ComponentType type) {
        componentBits |= type.bit;
    }

    void removeComponentType(ComponentType type) {
        componentBits &= ~type.bit;
    }

    void addSystemEnabledBit(long systemBit) {
        systemEnabledBits |= systemBit;
    }

    void removeSystemEnabledBit(long systemBit) {
        systemEnabledBits &= ~systemBit;
    }

    void reset() {
        componentBits = 0;
        systemEnabledBits = 0;
        active = false;
        removed = false;
        pendingRefresh = false;
        uniqueId = nextUniqueId++;
    }

    public String toString() {
        return "Entity[id=" + id + ", uniqueId=" + uniqueId + "]";
    }

}
